import java.util.Vector;
import java.util.*;
import java.awt.*;

/**
 * Each lattice site is assigned a host that can then contain different species (e.g. phage, bacteria)
 * @author Jan
 */
class Host {
    public Vector<Integer> limitedImV;
    Vector<Phage> phagevector;
    int x, y; //place of host in lattice
    //    Colormap cmapTW;
    double colIndex;
    Color hostColor; // - new ;
    int countDisease = 0;
    int presentDisease = -1;
    int hostID = -1;
    int maxImmunityM;
    int pop=0;
    Phage p;
    Bacteria b;
    int index;
    int2D[] neighbors;
    Host[] neighborHosts;
    int Nneighbors;
    
    
    public void reset(){
	pop=0;
	colIndex=0.;
	if (p!=null) p=null;
	if (b!=null) b=null;		
    }

    
    /**
     * contructor
     */
    public Host() {
    }
    
    
    /**
     * create new host / node
     * @param xx x-value in lattice
     * @param yy y-value in lattice
     * @param maxImmunity How many diseases host can be immune to. 
     */
    public void init(int xx, int yy, int maxImmunity, int Nneighbors, int2D[] neighbors) {
        int neighX=0,neighY=0;
	limitedImV = new Vector<Integer>(); //Vector with values host is immune to.
	phagevector= new Vector<Phage>();
        maxImmunityM = maxImmunity;
        hostColor = new Color(0, 255, 255); //host is white
        countDisease = 0;
        presentDisease = -1; //minus one means no disease
        x = xx;
        y = yy;
	pop=0;
	colIndex=0.;
	this.Nneighbors=Nneighbors;
	this.neighbors=neighbors;

    }

    public void setNeighbors(Host[] h){
	this.neighborHosts=h;
	if (p!=null) p.setNeighbors(h);
	if (b!=null) b.setNeighbors(h);
    }
    
    public int2D[] getNeighbors(){
	return this.neighbors;	
    }

    public void checkPopulation(){
	if (p!=null) if (p.populationStatus==0) p=null;
	if (b!=null) if (b.populationStatus==0) b=null;       
    }

    public void update(int index, double time){
	boolean update = true;
	if (index<100)  if (this.b!=null){if (this.b.bactID==index)  {b.update(time); update=false;}};
	if (index>=100) if (this.p!=null){if (this.p.phageID==index) {p.update(b,time); update=false;}};
	
	checkPopulation();
	checkColor();
    }

    public void checkColor(){
	double im=1.,im1=1.;
	double immFact,popFact,specFact;
	if (p!=null){
	    if (p.phageID==100) {
		this.hostColor=new Color(50+p.burstsize*50,0,0);		
		colIndex=0;
	    }
	    else {
		this.hostColor=new Color(200,200,200);
		colIndex=1;
	    }
	    //if (p.phageID==100) this.hostColor=new Color(0,(int)(1.0*p.populationStatus/p.maxPopulationStatus*255.),0); //green	   
	    //if (p.phageID==101) this.hostColor=new Color((int)(1.0*p.populationStatus/p.maxPopulationStatus*255.),(int)(1.0*p.populationStatus/p.maxPopulationStatus*255.),0); //yellow
	    //if (p.phageID==102) this.hostColor=new Color(255,0,(int)(1.0*p.populationStatus/p.maxPopulationStatus*255.)); //pink
	    //if (p.phageID==103) this.hostColor=new Color(100,100,(int)(1.0*p.populationStatus/p.maxPopulationStatus*255.)); //purple
	    //colIndex=p.phageID-100+1;
	    if (p.expression>=0.0)
		this.hostColor=new Color(0,(int)(255.*p.expression),(int)(255.*(1.-p.expression)));
	    if (p.waitingTime!=1.0 && p.waitingTime>0.0){
		this.hostColor=new Color(0,(int)(Math.min(255,(int)(p.waitingTime*50.))),0);
		this.hostColor=new Color(0,Math.min(255,80*this.p.rna),100);
		colIndex=1./p.waitingTime;
	    }
	    if (b!=null){
		//		System.out.println("b not zero");
		boolean immu=false;
		for (int i=0;i<8;i++) if (b.immunity.contains(i)) immu=true;
		if (immu) {
		    this.hostColor=new Color(0,0,255);
		    colIndex=-1;
		}
		else {
		    this.hostColor=new Color(0,0,0);
		    colIndex=-1.;
		}
	    }
	}
	else{
	    if (b!=null){
		int immu=-1;
		for (int i=0;i<1;i++) if (b.immunity.contains(i)) immu=b.immunity.indexOf(i);
		if (immu>=0) {
		    int dummy=immu*15;
		    this.hostColor=new Color(0,Math.min(dummy,255),Math.max(255-dummy,0));
		    colIndex=-1-immu;
		}
		else {
		    this.hostColor=new Color(0,0,0);
		    colIndex=-1;
		}
		if (b.nRes==0) {this.hostColor=new Color(0,0,0);}
		//		System.out.println("bs "+b.burstsize);
		if (b.nRes==1) {this.hostColor=new Color(0,50*b.burstsize,255-50*b.burstsize);}
		if (b.nRes==2) {this.hostColor=new Color(0,0,0);}

		//if (b.nRes>0) this.hostColor=new Color(0,Math.min((int)(25.*(b.nRes)),255),Math.max(255-(int)(25.*(b.nRes)),0));

		//System.out.println("waiting time and nRes " + b.waitingTime + ",  " + b.nRes);

		//this.hostColor=new Color(0,(int)(b.waitingTime*20.),0);

	    }
	    else{
		//		System.out.println("test1 "+x+ " " + y);
		this.hostColor=new Color(255,255,255); //host is light gray
		colIndex=3;
	    }
	}
    }


    public void addBacteria(double time, int bactID, double expLevel, double waitingTime, Vector<Integer> imm, BinarySearchTree t){
	//	System.out.println("adding bact");
	if (this.b==null){
	    if (!(this.p!=null && this.p.slumber)){ // bacterial debris on the site blocks occupation with new bacteria
		    if (bactID==0) this.b=new bactA();
		    if (bactID==1) this.b=new bactB();
		    if (bactID==2) this.b=new bactMute();
		    if (bactID==3) this.b=new bactSwitcher();
		    this.b.init(x,y,time, bactID, expLevel, waitingTime, 1,100,imm, Nneighbors,neighbors,neighborHosts,t);
		    checkPopulation();
		    checkColor();
		}
	}
    }
    
    public void addPhage(double time, double waitingTime, int phageID, BinarySearchTree t){
	if (!(this.p!=null && this.p.slumber) && !(this.p!=null && this.p.phageID==phageID)){ // bacterial debris on the site blocks occupation with new bacteria, don't add phage to sites where the same phage is already active	
	    if (this.p==null && this.b!=null){ // only sites without phages present can be invaded
		this.p=new Phage();
		
		this.p.init(x,y,time,waitingTime,phageID,this.b,Nneighbors,neighbors,neighborHosts,t);
		if (p.populationStatus==0) p=null;
		checkPopulation();
		checkColor();
	    }
	}
    }
    

    public void addPhageCountExp(double time, double waitingTime, int phageID, BinarySearchTree t,makeAverage phagExp, makeAverage phagAdd){
	if (!(this.p!=null && this.p.slumber) && !(this.p!=null && this.p.phageID==phageID)){ // bacterial debris on the site blocks occupation with new bacteria, don't add phage to sites where the same phage is already active	
	    if (this.p==null && this.b!=null){ // only sites without bacteria present can be invaded
		this.p=new PhageCountExp();
		phagAdd.addElement(1.0);
		this.p.init(x,y,time,waitingTime,phageID,this.b,Nneighbors,neighbors,neighborHosts,t,phagExp,phagAdd);
		if (p.populationStatus==0) p=null;
		checkPopulation();
		checkColor();
	    }
	    else phagAdd.addElement(0.);
	}
	else phagAdd.addElement(0.);
    }
    

    public void addPhageCountExpBurst(double time, double waitingTime, int phageID, BinarySearchTree t,makeAverage phagExp, makeAverage phagAdd,boolean burstMutePhage,int bs,double muteRate){
	if (!(this.p!=null && this.p.slumber) && !(this.p!=null && this.p.phageID==phageID)){ // bacterial debris on the site blocks occupation with new bacteria, don't add phage to sites where the same phage is already active	
	    if (this.p==null && this.b!=null){ // only sites without infected bacteria present can be invaded
		//		System.out.println("adding phage 1");
		this.p=new PhageCountExpBurst();
		//		System.out.println("adding phage ");
		phagAdd.addElement(1.0);
		this.p.init(x,y,time,waitingTime,phageID,this.b,Nneighbors,neighbors,neighborHosts,t,phagExp,phagAdd,burstMutePhage,bs,muteRate);
		if (p.populationStatus==0) p=null;
		checkPopulation();
		checkColor();
	    }
	    else phagAdd.addElement(0.);
	}
	else phagAdd.addElement(0.);
    }
    

    public void addMutePhage(double time, double waitingTime, int phageID, BinarySearchTree t, double expLev){
    }

    public void addMutePhage2(double time, double waitingTime, int phageID, BinarySearchTree t, double expLev){
    }

    public void addMutePhage3(double time, double waitingTime, int phageID, BinarySearchTree t, double expLev){
    }

    public void addMuteBacteriaStandard(double time, int bactID, double expLevel, double waitingTime, Vector<Integer> imm, BinarySearchTree t, double rb, int nRes){
    }

    public void addMute_nres_Bacteria(double time, int bactID, double expLevel, double waitingTime, Vector<Integer> imm, Vector<Integer> fam, BinarySearchTree t, double rb, int nRes,double immSwitchRate,double waitingTime0,double rPrime, int[][] attackSurvival,double cost){
    }

    public void addMuteBacteria(double time, int bactID, double expLevel, double waitingTime, Vector<Integer> imm, BinarySearchTree t, double rb, int nRes){
    }

    public void addMute_nres_BacteriaBurst(double time, int bactID, double expLevel, double waitingTime, Vector<Integer> imm, Vector<Integer> fam, BinarySearchTree t, double rb, int nRes, double immSwitchRate, double waitingTime0, double rPrime, int[][] attackSurvival,double cost,int bs){
    }

    /**
     * Makes host immune to disease
     * @param disease Disease host should be immune to.
     */
    public void makeImmune(int disease) {
        limitedImV.add(disease);
        if (limitedImV.size() >= maxImmunityM) {
            limitedImV.remove(0);
        }
    }

    
    /**
     * Removes immunity of disease to host
     * @param disease Disease host should no longer be immune to.
     */
    public boolean removeImmune(int disease) {
            return limitedImV.remove((Integer) disease);
    }
    
    
	/**
	 * Returns true if host is immune to disease
	 * @param disease
	 * @param limitedImmunity
	 * @return
	 */
    public boolean isImmune(int disease) {
            if (limitedImV.contains(disease))
                return true;
             else
                return false;
            }

    public void update(){
	
    }

    public void spread(){
	
    }

}
