// -----------------------------------------------------------------------------
// Template.java
// -----------------------------------------------------------------------------

/*
 * =============================================================================
 * Applet that simulates the spread of diseases on a lattice
 * By Ala (plus Martin, Ludvig and Jeppe)
 * =============================================================================
 */

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.Hashtable;
import java.util.Vector;
import javax.swing.*;


public class Phage extends JApplet implements ActionListener, AdjustmentListener, Runnable, MouseListener {

	private static final long serialVersionUID = 1L;
	
	//Constants: Dynamics
	public static int L = 300; //Size of lattic
    int delay = 2; //how 
	int maxImmunity = 2000;
	public int updateEveryM0 = L * L / 20;
    public int updateEveryM = 20*updateEveryM0; //How many updates before lattice is redrawn
    public int diseasePool = 10 * L;
    double alpha = 0.05 / (L * L); //how often new diseases appear
	
	//Constants: Heights, widths and sizes
	public static int nodeSize = 2; //length and height of one node
    public static int PlotWidth = 200, PlotHeight = 100; //height and width of diversity plot
    public static int x0 = 20, y0 = 20; //starting coordinates for lattice
    public static int panelX = x0 + L*nodeSize + 50, panelY = 40; //x and y coordinates for panel
    static final int appletHeight = L*nodeSize+2*x0;
    static final int appletWidth = L*nodeSize + PlotWidth + 200; //height and width of Applet
        
    //Initializing parameters
    boolean keepMoving;
    boolean periodicBC = false;
	int cmapRangeTW = diseasePool;
    int[] diseaseSizes = new int[diseasePool];
	private int diseasesAlive = 0; //Number of diseases alive right now.
    public Host[] host;
    UpdateTimeSeries updateTimeSeries1 = new UpdateTimeSeries();
	URL codeBase;
	
    //graphics
    private volatile Thread animator = null;
    Image background2D;
    Graphics2D bufferedGraphics2D;
    RenderingHints hints;
    BasicStroke line_stroke;
    Colormap cmapTW;
    
    GradientPaint redtowhite, redtowhite1;
    Color lblue, white, yellow, lyellow, lbluish, navyblue, black, golden, nColor[], beige, orange;
    Color dgreen, lgreen, red, lred, dgrey, green, mblack, lgrey;
    Color cbackground, cborder;
    Color ctext, cbuttontext;
       
    //Main buttons
    Button button_stop;
	Button button_periodicBC;
	
	//Scrollbars
	Scrollbar scroll_alpha;
	double alphaN = alpha*(L*L), alphaNMin = 0.001, alphaNMax = 10;
	int nBINS = 100;
	int alphaX, alphaY;
	Scrollbar scroll_time;
	
    
	/**
	 * Initializes the applet by invoking createGUI
	 */
    @Override
    public void init() {
        try {
            SwingUtilities.invokeAndWait(new Runnable() {
                public void run() {
                    codeBase = getCodeBase();
                    createGUI(); 
                }
            });
        } catch (Exception e) {
            System.err.println("createGUI didn't successfully complete");
        }
    }

    /**
     * makes graphical user interface, made by Ala
     */
    public void createGUI() {
        setSize(appletWidth, appletHeight);

        background2D = this.createImage(appletWidth, appletHeight);
        bufferedGraphics2D = (Graphics2D) background2D.getGraphics();
        
        //Define colours
        white = new Color(255, 255, 255);
        lred = new Color(255, 153, 0);
        lblue = new Color(190, 203, 255);
        red = new Color(255, 42, 42);
        navyblue = new Color(0, 81, 186);
        black = new Color(0, 0, 0);
        orange = new Color(232, 170, 63);
        dgrey = new Color(51, 51, 51);
        lgrey = new Color(150, 150, 150);
        lgreen = new Color(0, 128, 0);
        cbackground = white;

        line_stroke = new BasicStroke(2.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL); //???

        hints = new RenderingHints(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED);
        hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        hints.put(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_SPEED);
        hints.put(RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_DISABLE);
        hints.put(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
        hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        hints.put(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);
        hints.put(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
        hints.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);


        //Buttons
        alphaX =  panelX + 40;
        alphaY = panelY-4;
        //Button parameters
        int butWidth = 60;
        int butHeight = 25;
       
        // Add scroll bar Alpha
        bufferedGraphics2D.setFont(new Font("SansSerif", Font.BOLD, 10));
        bufferedGraphics2D.drawString("Alpha", panelX , panelY-4);
        
        scroll_alpha = new Scrollbar(Scrollbar.HORIZONTAL, 10, 3, 0, nBINS + 3);
        scroll_alpha.addAdjustmentListener(this);
        scroll_alpha.setBounds(panelX, panelY, 100, 15);
        scroll_alpha.setBackground(Color.white);
        add(scroll_alpha);
        int  iAlpha  = (int)((nBINS -1)*Math.log(alphaN/alphaNMin)/Math.log(alphaNMax/alphaNMin));
        scroll_alpha.setValue(iAlpha);
        updateAlphaText(bufferedGraphics2D, alphaX, alphaY);

        // Add scroll bar Update Speed
        panelY += 45;
        bufferedGraphics2D.drawString("Update Speed", panelX, panelY-4);
        scroll_time = new Scrollbar(Scrollbar.HORIZONTAL, 10, 3, 1, 103);
        scroll_time.addAdjustmentListener(this);
        scroll_time.setBounds(panelX, panelY, 100, 15);
        scroll_time.setBackground(Color.white);
        add(scroll_time);
        scroll_time.setValue(updateEveryM/updateEveryM0);
        updateEveryM = updateEveryM0 * scroll_time.getValue();
        
        //Add stop button
        panelY += 45;
        button_stop = new Button("Pause");
        button_stop.setBounds(panelX, panelY, butWidth, butHeight);
        button_stop.setBackground(orange);
        add(button_stop);
        button_stop.addActionListener(this);

        //add text 'choose between...'
        panelY += 2*butHeight;
        bufferedGraphics2D.setFont(new Font("SansSerif", Font.BOLD, 10));
        bufferedGraphics2D.drawString("Choose between periodic or absorbing ", panelX + butWidth + 5, panelY + 10);
        bufferedGraphics2D.drawString("boundary conditions", panelX + butWidth + 5, panelY + 20);
        
        //add periodicBC button
        button_periodicBC = new Button("Periodic");
        button_periodicBC.setBounds(panelX, panelY, butWidth, butHeight);
        button_periodicBC.setBackground(lblue);
        add(button_periodicBC);
        button_periodicBC.addActionListener(this);

        panelY += 4*butHeight;

        setLayout(null);
        addMouseListener(this);
        updateTimeSeries1.set_timeframe();
        cmapTW = new Colormap("0ryb8", cmapRangeTW);
        
        initLattice(bufferedGraphics2D);
        updateTimeSeries1.init_timeframe();
        bufferedGraphics2D.setBackground(black);
    }
    
    
    /**
     * Draws lattice with initial infection
     * @param g	Graphics2D
     */
    public void initLattice(Graphics2D g) {

        host = new Host[L * L];
        int x = 0, y = 0;

        for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
                x = x0 + nodeSize * j;
                y = y0 + nodeSize * i;

                host[i * L + j] = new Host(); //create host
                host[i * L + j].hostID = i * L + j; //name host
                host[i * L + j].init(x, y, maxImmunity); //initialize host
                
                infectHost(i * L + j, 0); // initial infection
                g.setColor(host[i * L + j].hostColor);
                g.fillRect(x, y, nodeSize, nodeSize);
            }
        }
    }
    
    
    @Override
    /**
     * At every update updateEveryM timesteps are taken, lattice is redrawn and plot is updated
     */
    public void update(Graphics g) {

        for (int i = 0; i < this.updateEveryM; ++i) {
            newDisease();
            spreadToNeighbours();
        }

        bufferedGraphics2D.setStroke(new BasicStroke(2));
        //Divesity vs time
        draw_timeseries(bufferedGraphics2D, panelX, panelY, PlotWidth, PlotHeight, true, "Number of diseases"); 
        
        redrawLattice(bufferedGraphics2D);
      
        Graphics2D g2d = (Graphics2D) g;
        g2d.drawImage(background2D, 0, 0, this);
    }
    

    /**
     * Creates new desease with probability alpha
     */
    public void newDisease() {
    	int countTries = 0;
    	int rDis = 0;
    	
        if (Math.random() < alpha){ //infect random host with probability alpha
            while (countTries<10) { //assign a random rDis to disease
                rDis = (int) (1000 * (diseasePool) * Math.random()) % diseasePool; //row of disease
                if (diseaseSizes[rDis] == 0){ //make sure it is a new disease
                  break;
                }
                countTries++;
            }
            int rHost = (int) (1000 * L * L * Math.random()) % (L * L); //row of host to get infected.
            infectHost(rHost, rDis);
        }
    }

    
    /**
     * Spreads a disease from a random node to random neighbour with periodic boundary conditions
     */
    public void spreadToNeighbours() {
        int rHost = (int) ((L * L) * Math.random()); //row of host to get infected
        int i = rHost / L; //x-coordinate of host to get infected
        int j = rHost % L; //y-coordinate of host to get infected
        int in, jn; //x- and y-coordinate of host to spread disease
        int rn = (int) (4 * Math.random()); //direction get disease from
        
        if(periodicBC){
            //find coordinates (in,jn) of host that should infect
            if (rn == 0) { 
                in = (L + i - 1) % L;
                jn = j;
            } else if (rn == 1) {
                in = i;
                jn = (L + j - 1) % L;
            } else if (rn == 2) {
                in = (L + i + 1) % L;
                jn = j;
            } else{
                in = i;
                jn = (L + j + 1) % L;
            }
        } else{ //if the boundaries are absorbing
        	if (rn == 0) {
                in = i - 1;
                jn = j;
            } else if (rn == 1) {
                in = i;
                jn = j - 1;
            } else if (rn == 2) {
                in =  i + 1;
                jn = j;
            } else{
                in = i;
                jn = j + 1;
            }
        	if((in<0)||(jn<0)||(in>=L)||(jn>=L)) { //if host is outside lattice, return.
        		return;
        	}
        }
        
        int rInfectingHost = in * L + jn; //row of host 
        int dis = host[rInfectingHost].presentDisease; //transmitted disease
        
        if (!host[rHost].isImmune(dis)) {//if host is not immune
            infectHost(rHost, dis); //infect host with disease
        }
    }

    
    /**
     * removes 'disease' from all nodes.
     * @param disease
     */
    public void eraseDiseaseFromAllHostsImmunity(int disease) {
        for (int i = 0; i < L * L; ++i) {
            if (host[i].removeImmune(disease)) {
                host[i].countDisease--;
            }
        }
    }

    
    /**
     * Infects a host with a disease
     * @param rHost host to get infected
     * @param disease disease that host should be infected with.
     */
    public void infectHost(int rHost, int disease) {
        int oldDis = host[rHost].presentDisease; //old disease node had.
        	
    	if (oldDis != -1) { //if node was ill beforehand
            diseaseSizes[oldDis]--; //decrease the size of old disease
            
            if( diseaseSizes[oldDis] ==0){ //if old disease is distinct:
            	diseasesAlive--; //record this.
            	eraseDiseaseFromAllHostsImmunity(oldDis); //erase immunity
            }
        }
        
        host[rHost].presentDisease = disease; //overwrite old disease with new. 
        host[rHost].makeImmune(disease); // make host immune to new disease.
        host[rHost].countDisease++; //add one to total number of diseasees host has had
        
        if(diseaseSizes[disease]==0){ //if this was a new disease, add one to the total number of living diseases
        	diseasesAlive++;
        }
        
        diseaseSizes[disease]++; //increase size of disease by 1.
        host[rHost].hostColor = cmapTW.col(disease); //change colour of host to disease colour.
    }
    
    
    /**
     * Draws the lattice
     * @param g Graphics2D
     */
    public void redrawLattice(Graphics2D g) {
        int x, y;

        for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
                x = x0 + nodeSize * j;
                y = y0 + nodeSize * i;
                g.setColor(host[i * L + j].hostColor);
                g.fillRect(x, y, nodeSize, nodeSize);
            }
        }
    }

    
    /** 
     * Returns an ImageIcon, or null if the path was invalid. 
     */
    protected ImageIcon createImageIcon(String path) {
        java.net.URL imgURL = getClass().getResource(path);
        if (imgURL != null) {
            return new ImageIcon(imgURL);
        } else {
            System.err.println("Couldn't find file: " + path);
            return null;
        }
    }

    
    @Override
    public void start() {
        animator = new Thread(this);
        animator.start();
        keepMoving = true;
    }

    @Override
    public void stop() {
        animator = null;
        keepMoving = false;

    }
    
    @Override
    public void destroy() {
        removeAll();
    }

    public void run() {
        Thread thisThread = Thread.currentThread();
        while (animator == thisThread) {
            try {
                Thread.sleep(delay);
            } catch (InterruptedException e){}
            if(keepMoving) {
                repaint();
            }
        }
    }

    @Override
    public void paint(Graphics g) {
        update(this.getGraphics());

    }

    public void mouseClicked(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    /**
     * Updates the shown value of alpha in applet.
     */
    public void updateAlphaText(Graphics2D g, int textOut_x, int textOut_y){
         g.setColor(cbackground);
         g.fillRect(textOut_x-5, textOut_y-10, 125, 80);// to cover text
         g.setColor(dgrey);
         g.setFont(new Font("Helvetica", Font.BOLD, 10));
         g.drawString("" +  String.format("%1$4.3f",  alphaN), textOut_x, textOut_y);

    }

    
    /**
     * Is activated if user adjusts alphavalue via scrollbar
     */
    public void adjustmentValueChanged(AdjustmentEvent e) {
        if (e.getSource() == scroll_alpha) {
            alphaN = alphaNMin * Math.pow(alphaNMax / alphaNMin, 1.0 * scroll_alpha.getValue() / nBINS);
            alpha = alphaN / (L * L);
            
            updateAlphaText(bufferedGraphics2D, alphaX, alphaY);
        }
        if (e.getSource() == scroll_time) {
            updateEveryM = updateEveryM0 * scroll_time.getValue();
        }
        keepMoving = true;
    }
    
    
    /**
     * Inherited method that manages actions done by user.
     */
    public void actionPerformed(ActionEvent e) {
       
    	//Manage periodic button
        if (e.getSource() == button_periodicBC) {
           if(!periodicBC) {
	           periodicBC = true;
	           button_periodicBC.setLabel("Absorbing");
	           button_periodicBC.setBackground(orange);
           }
           else {
	           periodicBC = false;
	           button_periodicBC.setLabel("Periodic");
	           button_periodicBC.setBackground(lblue);
           }
        }
        
        //Manage stop button
        if (e.getSource() == button_stop) { 
            if (keepMoving) { //if 'stop' is pressed
                keepMoving = false; //stop updating
                animator = null;
                button_stop.setLabel("Run");
                button_stop.setBackground(lred);
            } else { //if 'run' is pressed
                keepMoving = true;
                animator = new Thread(this);
                animator.start();
                button_stop.setLabel("Pause");
            }
        }
    }


    /**
     * Draws Divesity vs time
     * @param g
     * @param plotX	X-value of plot
     * @param plotY	Y-value of plot
     * @param plotW	Width of plot
     * @param plotH	Height of plot
     * @param xTics 'true' if x-tics should be drawn
     * @param title
     */
    public void draw_timeseries(Graphics2D g, int plotX, int plotY, int plotW, int plotH, boolean xTics, String title) {

        g.finalize();
        
        //Draw title:
        g.setColor(dgrey);
        g.setFont(new Font("Helvetica", Font.BOLD, 17));
        g.drawString(title, plotX + plotW / 2 - g.getFontMetrics().stringWidth(title) / 2, plotY - 5);
        
        //Draw background of plot
        g.setColor(cbackground);
        g.fillRect(plotX - 40, plotY, plotW + 60, plotH + 15);
        
        //Draw axes and tics
        g.setColor(dgrey);
        // y axis
        g.fillRect(plotX, plotY + plotH, plotW, 2);
        // x axis
        g.fillRect(plotX, plotY, 2, plotH + 2);
        //ytics
        g.fillRect(plotX - 3, plotY + 20, 4, 2);
        g.fillRect(plotX - 3, plotY + plotH, 4, 2);
        //xtics
        g.fillRect(plotX, plotY + plotH, 2, 4);
        g.fillRect(plotX + plotW - 2, plotY + plotH, 2, 4);
        
        //Draw labels
        if (xTics) {
            draw_XTicLabels(g, plotX, plotY, plotW, plotH);
        }
        g.setColor(dgrey);
        g.setStroke(line_stroke);
        double MaxY = updateTimeSeries1.recheckMaxY();
        double MinY = updateTimeSeries1.recheckMinY();
        draw_YTicLabels(g, plotX, plotY, plotH, MaxY, MinY);
        
        //draw plot on the fly
        updateTimeSeries1.updateTimeSeries_l(diseasesAlive-1, MaxY, MinY, plotX, plotY, plotW, plotH);
        bufferedGraphics2D.drawPolyline(updateTimeSeries1.dtime, updateTimeSeries1.TimeSeries, updateTimeSeries1.Plottime - 1);
    }

    
    /**
     * Draws the labels on the y-axis of diversity-graph
     * @param g
     * @param plotX	X-value of diversity plot
     * @param plotY	Y-value of diversity plot
     * @param plotH	Height of diversity plot
     * @param MaxY 	The maximum diversity in timeseries (upper label)
     * @param MinY	The minimum diversity in timeseries (lower label)
     */
    public void draw_YTicLabels(Graphics2D g, int plotX, int plotY, int plotH, double MaxY, double MinY) {
        g.setColor(dgrey);

        g.setFont(new Font("SansSerif", Font.BOLD, 11));
        g.drawString(String.format("%1$2.0f", MaxY), plotX - 35, plotY + 25);
        g.drawString(String.format("%1$2.0f", MinY), plotX - 35, plotY + plotH);

    }
    
    
    /**
     * Draws the labels on the x-axis of diversity-plot (NB: Does not take into account that update speed can change)
     * @param g
     * @param plotX
     * @param plotY
     * @param plotW
     * @param plotH
     */
    public void draw_XTicLabels(Graphics2D g, int plotX, int plotY, int plotW, int plotH) {
    
        g.setColor(dgrey);
        int tMaxM = (int) (updateTimeSeries1.time_frame * this.updateEveryM/(L*L));
        String tMin = String.format("%1$3d", 0);
        String t1_4 = String.format("%1$3d", tMaxM / 4);
        String t1_2 = String.format("%1$3d", tMaxM / 2);
        String t3_4 = String.format("%1$3d", 3 * tMaxM / 4);
        String tMax = String.format("%1$3d", tMaxM);

        g.setFont(new Font("SansSerif", Font.BOLD, 11));
        int xx = plotX + plotW - g.getFontMetrics().stringWidth(tMax) / 2;
        int xx1_4 = (int) (plotX + plotW / 4 - g.getFontMetrics().stringWidth(t1_4) / 2);
        int xx1_2 = (int) (plotX + plotW / 2 - g.getFontMetrics().stringWidth(t1_2) / 2);
        int xx3_4 = (int) (plotX + 3 * plotW / 4 - g.getFontMetrics().stringWidth(t3_4) / 2);
        int yy = plotY + plotH + 13;

        g.drawString(tMin, plotX - 8, yy);
        g.drawString(t1_4, xx1_4, yy);
        g.drawString(t1_2, xx1_2, yy);
        g.drawString(t3_4, xx3_4, yy);
        g.drawString(tMax, xx, yy);
    }
}


class UpdateTimeSeries {
    public int time_frame = 6000;
    public int maxtime_frame = 10000;
    public int Plottime;
    public int TimeSeries[], dtime[];
    public double TimeSeries_raw[];
    public int Ptime;

    
    public void set_timeframe() {
        TimeSeries = new int[this.maxtime_frame];
        TimeSeries_raw = new double[this.maxtime_frame];
        dtime = new int[this.time_frame];
    }

    
    public int getPtime() {
        return this.Ptime;
    }
    

    public void init_timeframe() {
        for (int i = 0; i < maxtime_frame; ++i) {
            TimeSeries_raw[i] = 0;
        }
        Ptime = 0;
        Plottime = 0;
    }

    
    /**
     * returns the maximum value in TimeSeries_raw
     */
    public double recheckMaxY() {
        double maxY = 0;
        for (int i = 0; i < this.Ptime; i++) {
            if (TimeSeries_raw[i] > maxY) {
                maxY = TimeSeries_raw[i];
            }
        }
        return maxY;
    }

    
    /**
     * returns the minimum value in TimeSeries_raw
     */
    public double recheckMinY() {
        double minY = 100000;
        for (int i = 0; i < this.Ptime; i++) {
            if (TimeSeries_raw[i] < minY) {
                minY = TimeSeries_raw[i];
            }
        }
        return minY;
    }


    /**
     * Updates the diversity plot on the fly
     * @param y				count of diseases alive rigth now
     * @param MaxY			Max diversity in timeseries
     * @param MinY			Min diversity in timeseries
     * @param PlotX			X-coordinate of plot
     * @param PlotY			Y-coordinate of plot
     * @param PlotWidth		Width of plot
     * @param PlotHeight	Height of plot
     */
    public void updateTimeSeries_l(double y, double MaxY, double MinY, int PlotX, int PlotY, int PlotWidth, int PlotHeight) {

        if (this.Ptime < this.time_frame) {
            for (int i = 0; i < this.Ptime; i++) {
                dtime[this.Ptime - i] = dtime[Ptime - i - 1];
                TimeSeries_raw[Ptime - i] = TimeSeries_raw[Ptime - i - 1];

            }
            dtime[0] = this.Ptime * PlotWidth / time_frame + PlotX;
            this.Ptime++;
        } else {
            for (int i = maxtime_frame - 1; i > 0; i--) {
                TimeSeries_raw[i] = TimeSeries_raw[i - 1];
            }
        }

        TimeSeries_raw[0] = y;

        if (MinY > 0.9 * MaxY) {
            MinY = 0.9 * MaxY;
        }
        for (int i = 0; i < this.Ptime; ++i) {
            TimeSeries[i] = PlotY - (int) ((TimeSeries_raw[i] - MinY) / (MaxY - MinY) * TimeSeries_raw[i] / MaxY * (PlotHeight - 20)) + PlotHeight - 5;
            if (TimeSeries[i] > PlotY + PlotHeight + 5) {
                TimeSeries[i] = PlotY + PlotHeight + 5;
            }
            if (TimeSeries[i] < PlotY) {
                TimeSeries[i] = PlotY;
            }
        }
        if (this.Ptime < this.time_frame - 1) {
            this.Plottime = this.Ptime;
        } else {
            this.Plottime = this.time_frame;
        }
    }
}


/**
 * Each host is placed at a node in the lattice, and is carrier of 0-1 disease. 
 * @author Jeppe
 */
class Host {
    public Vector<Integer> limitedImV;

    int x, y; //place of host in lattice
    Color hostColor; // - new ;
    int countDisease = 0;
    int presentDisease = -1;
    int hostID = -1;
    int maxImmunityM;

    
    /**
     * contructor
     */
    public Host() {
    }
    
    
    /**
     * create new host / node
     * @param xx x-value in lattice
     * @param yy y-value in lattice
     * @param maxImmunity How many diseases host can be immune to. 
     */
    public void init(int xx, int yy, int maxImmunity) {
        limitedImV = new Vector<Integer>(); //Vector with values host is immune to.
        maxImmunityM = maxImmunity;
        hostColor = new Color(255, 255, 255); //host is white
        countDisease = 0;
        presentDisease = -1; //minus one means no disease
        x = xx;
        y = yy;
    }
    
    
    /**
     * Makes host immune to disease
     * @param disease Disease host should be immune to.
     */
    public void makeImmune(int disease) {
        limitedImV.add(disease);
        if (limitedImV.size() >= maxImmunityM) {
            limitedImV.remove(0);
        }
    }

    
    /**
     * Removes immunity of disease to host
     * @param disease Disease host should no longer be immune to.
     */
    public boolean removeImmune(int disease) {
            return limitedImV.remove((Integer) disease);
    }
    
    
	/**
	 * Returns true if host is immune to disease
	 * @param disease
	 * @param limitedImmunity
	 * @return
	 */
    public boolean isImmune(int disease) {
            if (limitedImV.contains(disease))
                return true;
             else
                return false;
            }
    }


class Colormap {

    private int max_colors;			// maximum numbers of colors available
    private String range_string;	// color range "BRYW" for Black, Yellow, etc
    private int interval_count;		// color range - how many colors?
    private int interval_size;		// how many colors per interval?
    private int interval_rgb[][];	// rgb value of each interval
    public Color table[];			// colormap table

    // setup interval rgb table
    // interval_count needs to be set up before entering this function
    private void setup_interval_rgb(String spectrum_str) {
    	Hashtable<Character, Integer> color_to_int = new Hashtable<Character, Integer>();
        interval_rgb = new int[interval_count + 1][3];

        // setup rgb values for each numbered colors
        int int_to_rgb[][] = {
            {255, 0, 0}, // 0  : r = Red
            {0, 255, 0}, // 1  : g = Green
            {0, 0, 255}, // 2  : b = Blue
            {255, 255, 0}, // 3  : y = Yellow
            {0, 255, 255}, // 4  : c = Cyan
            {255, 0, 255}, // 5  : m = Magenta
            {0, 0, 0}, // 6  : 0 = Black
            {255, 255, 255}, // 7  : 1 = White
            {120, 72, 0}, // 8  : 2 = brown   Palette: the persistence of memory (S. Dali)
            {216, 216, 96}, // 9  : 3 = brown yellow
            {96, 168, 168}, // 10 : 4 = blue
            {72, 48, 0}, // 11 : 5 = Dark brown
            {144, 192, 168}, // 12 : 6 = blue-green
            {11, 17, 13}, // 13 : 7 = Black  Palette: the scream (E. Munch)
            {44, 77, 86}, // 14 : 8 = Blue-black
            {195, 170, 114}, // 15 : 9 = Beige-brown
            {220, 118, 18}, // 16 : 10 = Orange-red
            {189, 50, 0} // 17 : 11 = Red
        };
        // you can add more RGB values here and
        // set up the index below

        // setup color_to_int hashtable
        color_to_int.put(new Character('r'), new Integer(0));
        color_to_int.put(new Character('g'), new Integer(1));
        color_to_int.put(new Character('b'), new Integer(2));
        color_to_int.put(new Character('y'), new Integer(3));
        color_to_int.put(new Character('c'), new Integer(4));
        color_to_int.put(new Character('m'), new Integer(5));
        color_to_int.put(new Character('0'), new Integer(6)); // '0' means Black
        color_to_int.put(new Character('1'), new Integer(7)); // '1' means White
        color_to_int.put(new Character('2'), new Integer(8));  // Dhali
        color_to_int.put(new Character('3'), new Integer(9));
        color_to_int.put(new Character('4'), new Integer(10));
        color_to_int.put(new Character('5'), new Integer(11));
        color_to_int.put(new Character('6'), new Integer(12));
        color_to_int.put(new Character('7'), new Integer(13));  // Munch
        color_to_int.put(new Character('8'), new Integer(14));
        color_to_int.put(new Character('9'), new Integer(15));
        color_to_int.put(new Character('O'), new Integer(16));
        color_to_int.put(new Character('R'), new Integer(17));


        // fill in the rgb values for each interval
        for (int i = 0; i < interval_count + 1; ++i) {
            for (int j = 0; j < 3; ++j) {
                Integer idx = (Integer) color_to_int.get(
                        new Character(spectrum_str.charAt(i)));
                interval_rgb[i][j] = int_to_rgb[idx.intValue()][j];
            }
        }
    }

    
    /**
     * Constructor. 
     * spectrum_str: 	color spectrum - string of colours. "0ry1" for black to Red to Yellow to White.
     * max: 			max color range
     */
    public Colormap(String spectrum_str, int max) {
        int count = 0;

        // constants and parameters
        max_colors = max;
        range_string = spectrum_str;
        interval_count = range_string.length() - 1;
        interval_size = max_colors / interval_count;

        // setup rgb table for each interval
        this.setup_interval_rgb(range_string);

        // Allocate table[] and assign colors
        table = new Color[max_colors];
        for (int i = 0; i < interval_count; ++i) {
            // Get the start color in this range
            int start_range = i * interval_size;
            int end_range = (i == interval_count - 1)
                    ? max_colors
                    : (i + 1) * interval_size;
            int range_size = end_range - start_range;

            // for each rgb, find start and end and step
            float rstart = interval_rgb[i][0];
            float rend = interval_rgb[i + 1][0];
            float rstep = (rstart - rend == 0) ? 0 : -(rstart - rend) / range_size;
            float gstart = interval_rgb[i][1];
            float gend = interval_rgb[i + 1][1];
            float gstep = (gstart - gend == 0) ? 0 : -(gstart - gend) / range_size;
            float bstart = interval_rgb[i][2];
            float bend = interval_rgb[i + 1][2];
            float bstep = (bstart - bend == 0) ? 0 : -(bstart - bend) / range_size;

    
            for (int j = start_range; j < end_range; ++j) {
                table[count++] = new Color(
                        (int) (rstart + (j - start_range) * rstep),
                        (int) (gstart + (j - start_range) * gstep),
                        (int) (bstart + (j - start_range) * bstep));
            }
        }
    }

    // Return a color in the palette: 0<element<max_colors
    public Color col(int element) {
        return table[element];
    }
}

